/*
 * version.c
 *
 * Copyright (C) 1989, 1991, Craig E. Kolb
 * All rights reserved.
 *
 * This software may be freely copied, modified, and redistributed
 * provided that this copyright notice is preserved on all copies.
 *
 * You may not distribute this software, in whole or in part, as part of
 * any commercial product without the express consent of the authors.
 *
 * There is no warranty or other guarantee of fitness of this software
 * for any purpose.  It is provided solely "as is".
 *
 * version.c,v 1.1.1.1 1995/02/27 07:38:48 explorer Exp
 *
 * version.c,v
 * Revision 1.1.1.1  1995/02/27 07:38:48  explorer
 * Import of Netshade
 *
 * Revision 1.1.1.1  1994/11/27  03:55:48  explorer
 * Initial import
 *
 * Revision 4.1  1994/08/09  08:06:28  explorer
 * Bump version to 4.1
 *
 * Revision 1.1.1.1  1994/08/08  04:52:27  explorer
 * Initial import.  This is a prerelease of 4.0.6enh3, or 4.1 possibly.
 *
 * Revision 4.0  91/07/17  17:37:17  kolb
 * Initial version.
 * 
 */
#include "rayshade.h"
#include "patchlevel.h"
#include "stats.h"

void
VersionPrint()
{
	extern char rcsid[];

	fprintf(Stats.fstats,
		"raypaint: %s\nPatch level %d\n", rcsid, PATCHLEVEL);
}
