/* spot.h,v 1.1.1.1 1995/02/27 07:38:30 explorer Exp */

/*
 * Copyright (C) 1989, 1991, Craig E. Kolb
 * All rights reserved.
 *
 * This software may be freely copied, modified, and redistributed
 * provided that this copyright notice is preserved on all copies.
 *
 * You may not distribute this software, in whole or in part, as part of
 * any commercial product without the express consent of the authors.
 *
 * There is no warranty or other guarantee of fitness of this software
 * for any purpose.  It is provided solely "as is".
 *
 */

#ifndef SPOT_H
#define SPOT_H

#define LightSpotCreate(c,f,t,p,i,o) LightCreate( \
			(LightRef)SpotCreate(f,t,p,i,o), SpotMethods(), c)
typedef struct {
  Vector pos, dir;
  Float coef, radius, falloff;
} Spotlight;

extern Spotlight *SpotCreate _PROTO((Vector *, Vector *, Float, Float, Float));
extern LightMethods *SpotMethods _PROTO((void));
extern int SpotIntens _PROTO((Spotlight *, Color *, ShadowCache *, Ray *,
			      Float, int, Color *));
extern void SpotDirection _PROTO((Spotlight *, Vector *, Vector *, Float *));
extern void SpotMethodRegister _PROTO((UserMethodType));

#endif /* SPOT_H */
