/* windy.h,v 1.1.1.1 1995/02/27 07:38:43 explorer Exp */

/*
 * Copyright (C) 1989, 1991, Craig E. Kolb
 * All rights reserved.
 *
 * This software may be freely copied, modified, and redistributed
 * provided that this copyright notice is preserved on all copies.
 *
 * You may not distribute this software, in whole or in part, as part of
 * any commercial product without the express consent of the authors.
 *
 * There is no warranty or other guarantee of fitness of this software
 * for any purpose.  It is provided solely "as is".
 *
 */

#ifndef WINDY_H
#define WINDY_H

#define TextWindyCreate(s,w,c,b,o,t,h,i) TextCreate( \
		(TextRef)WindyCreate(s,w,c,b,o,t,h,i), WindyApply)

typedef struct {
  Float	scale;
  Float	windscale;
  Float	chaoscale;
  Float	bumpscale;
  Float	tscale;
  Float	hscale;
  Float	offset;
  int	octaves;
} WindyText;

extern WindyText *WindyCreate _PROTO((Float, Float, Float, Float, int,
				      Float, Float, Float));
extern void WindyApply _PROTO((WindyText *, Geom *, Ray *, Vector *,
			       Vector *, Vector *, Surface *));

#endif /* WINDY_H */
