/* bump.h,v 1.1.1.1 1995/02/27 07:38:41 explorer Exp */

/*
 * Copyright (C) 1989, 1991, Craig E. Kolb
 * All rights reserved.
 *
 * This software may be freely copied, modified, and redistributed
 * provided that this copyright notice is preserved on all copies.
 *
 * You may not distribute this software, in whole or in part, as part of
 * any commercial product without the express consent of the authors.
 *
 * There is no warranty or other guarantee of fitness of this software
 * for any purpose.  It is provided solely "as is".
 *
 */

#ifndef BUMP_H
#define BUMP_H

#define TextBumpCreate(s)	TextCreate((TextRef) BumpCreate(s), \
					BumpApply)
typedef struct {
  Float size;
} Bump;

extern Bump *BumpCreate _PROTO((Float));
extern void BumpApply _PROTO((Bump *, Geom *, Ray *, Vector *, Vector *,
			      Vector *, Surface *));

#endif /* BUMP_H */
