/* priority.c,v 1.1.1.1 1995/02/27 07:38:52 explorer Exp */

#ifdef __GNUC__
#include <inline/exec.h>
#else
#include <proto/exec.h>
#endif

long oldpri = 0;
int HavePri = 0;

void
ReducePriority(void)
{
  oldpri = SetTaskPri(FindTask(0L), -1);
  HavePri = 1;
}

void
RestorePriority(void)
{
  if (HavePri) {
    SetTaskPri(FindTask(0L), oldpri);
  }
}
