/* fog.h,v 1.1.1.1 1995/02/27 07:38:39 explorer Exp */

/*
 * Copyright (C) 1989, 1991, Craig E. Kolb
 * All rights reserved.
 *
 * This software may be freely copied, modified, and redistributed
 * provided that this copyright notice is preserved on all copies.
 *
 * You may not distribute this software, in whole or in part, as part of
 * any commercial product without the express consent of the authors.
 *
 * There is no warranty or other guarantee of fitness of this software
 * for any purpose.  It is provided solely "as is".
 *
 */

#ifndef FOG_H
#define FOG_H

#define AtmosFogCreate(c,t)	AtmosCreate((AtmosRef) FogCreate(c,t), \
					     FogApply)	
typedef struct Fog {
  Color	color;		/* Fog color */
  Color	trans;		/* Fog transmissivity */
} Fog;

extern Fog *FogCreate _PROTO((Color *, Color *));
extern void FogApply _PROTO((Fog *, Ray *, Vector *, Float, Color *));

#endif /* FOG_H */
