/* strerror.c,v 1.1 1995/08/17 23:23:12 explorer Exp */

#include <stdlib.h>
#include <string.h>

/*
 * replacement function for strerror.  This is a quick hack, but will
 * allow those missing strerror() to have SOMETHING at least.
 */

static char strerror_errbuf[40];

char *
strerror(errnum)
     int errnum;
{
  sprintf(strerror_errbuf, "Error number %d", errnum);
  return strerror_errbuf;
}

