/* fl.h,v 1.1.1.1 1995/02/27 07:38:32 explorer Exp */

/*
 * 	Copyright (C) 1992, Lawrence K. Coffin.
 *	Parts Copyright (C) 1989, 1991, Craig E. Kolb.
 *	All rights reserved.
 * 
 *	Parts of this program are from "Numerical Recipes in C" by William H.
 *	Press, Brian P. Flannery, Saul A. Teukolsky, and William T. Vetterling.
 *
 *	This software may be freely copied, modified, and redistributed
 *	provided that this copyright notice is preserved on all copies.
 *
 *	You may not distribute this software, in whole or in part, as part of
 *	any commercial product without the express consent of the authors.
 * 
 *	There is no warranty or other guarantee of fitness of this software
 *	for any purpose.  It is provided solely "as is".
 *
 */

#ifndef FL_H
#define FL_H

#define GeomFlCreate(f,g) GeomCreate((GeomRef)FlCreate(f,g), FlMethods())

extern GeomRef FlCreate _PROTO((long, int));
extern Methods *FlMethods METHODS_PROTO;

#endif /* FL_H */
