/* allocmatrix.h,v 1.1.1.1 1995/02/27 07:38:30 explorer Exp */

/*
 * Matrix allocation and freeing Functions.
 *
 * allocmatrix.h
 *
 * Peter Janssen - janssen@cs.kuleuven.ac.be
 *
 */

extern voidstar *allocMatrix _PROTO((size_t, size_t, size_t));
extern void freeMatrix _PROTO((voidstar *, size_t));

#define AllocMatrix(object, y, x)   \
    ((object **) allocMatrix(sizeof(object), y, x))
#define FreeMatrix(matrix, y) freeMatrix((voidstar *) matrix, y)
