/* gloss.h,v 1.1.1.1 1995/02/27 07:38:41 explorer Exp */

/*
 * Copyright (C) 1989, 1991, Craig E. Kolb
 * All rights reserved.
 *
 * This software may be freely copied, modified, and redistributed
 * provided that this copyright notice is preserved on all copies.
 *
 * You may not distribute this software, in whole or in part, as part of
 * any commercial product without the express consent of the authors.
 *
 * There is no warranty or other guarantee of fitness of this software
 * for any purpose.  It is provided solely "as is".
 *
 */

#ifndef GLOSS_H
#define GLOSS_H

#define TextGlossCreate(g)	TextCreate((TextRef) GlossCreate(g), \
					GlossApply)
typedef struct {
  Float glossy;		/* glossiness */
} Gloss;

extern Gloss *GlossCreate _PROTO((Float));
extern void GlossApply _PROTO((Gloss *, Geom *, Ray *, Vector *, Vector *,
			       Vector *, Surface *));

#endif /* GLOSS_H */
