/* blotch.h,v 1.1.1.1 1995/02/27 07:38:40 explorer Exp */

/*
 * Copyright (C) 1989, 1991, Craig E. Kolb
 * All rights reserved.
 *
 * This software may be freely copied, modified, and redistributed
 * provided that this copyright notice is preserved on all copies.
 *
 * You may not distribute this software, in whole or in part, as part of
 * any commercial product without the express consent of the authors.
 *
 * There is no warranty or other guarantee of fitness of this software
 * for any purpose.  It is provided solely "as is".
 *
 */

#ifndef BLOTCH_H
#define BLOTCH_H

#define TextBlotchCreate(m,s)	TextCreate((TextRef) BlotchCreate(m,s), \
					BlotchApply)
typedef struct {
  Float mix;
  Surface *surf;
} Blotch;

extern Blotch *BlotchCreate _PROTO((Float, Surface *));
extern void BlotchApply _PROTO((Blotch *, Geom *, Ray *, Vector *,
				Vector *, Vector *, Surface *));

#endif /* BLOTCH_H */
