/* builtin.h,v 1.2 1995/04/08 19:53:01 explorer Exp */

/*
 * Copyright (C) 1989, 1991, Craig E. Kolb
 * All rights reserved.
 *
 * This software may be freely copied, modified, and redistributed
 * provided that this copyright notice is preserved on all copies.
 *
 * You may not distribute this software, in whole or in part, as part of
 * any commercial product without the express consent of the authors.
 * 
 * There is no warranty or other guarantee of fitness of this software
 * for any purpose.  It is provided solely "as is".
 *
 */

#ifndef BUILTIN_H
#define BUILTIN_H

#include "libcommon/common.h"

extern Float SumExpr _PROTO((Float, Float));
extern Float DiffExpr _PROTO((Float, Float));
extern Float MultExpr _PROTO((Float, Float));
extern Float DivideExpr _PROTO((Float, Float));
extern Float ModExpr _PROTO((Float, Float));
extern Float NegateExpr _PROTO((Float));
extern Float LinearTime _PROTO((Float, Float, Float, Float));
extern Float CondExpr _PROTO((Float, Float, Float));
extern Float EqualExpr _PROTO((Float, Float));
extern Float UnequalExpr _PROTO((Float, Float));
extern Float LessExpr _PROTO((Float, Float));
extern Float LessEqualExpr _PROTO((Float, Float));
extern Float GreaterExpr _PROTO((Float, Float));
extern Float GreaterEqualExpr _PROTO((Float, Float));
extern Float AndExpr _PROTO((Float, Float));
extern Float OrExpr _PROTO((Float, Float));

extern Float LinearMode _PROTO((Float));
extern Float AccelMode _PROTO((Float));
extern Float DecelMode _PROTO((Float));
extern Float AccelDecelMode _PROTO((Float));
extern Float JumpMode _PROTO((Float));
  
#endif /* BUILTIN_H */
