/* roots.h,v 1.1.1.1 1995/02/27 07:38:37 explorer Exp */

/*
 * Copyright (C) 1989, 1991, Craig E. Kolb
 * All rights reserved.
 *
 * This software may be freely copied, modified, and redistributed
 * provided that this copyright notice is preserved on all copies.
 *
 * You may not distribute this software, in whole or in part, as part of
 * any commercial product without the express consent of the authors.
 *
 * There is no warranty or other guarantee of fitness of this software
 * for any purpose.  It is provided solely "as is".
 *
 */

#ifndef ROOTS_H
#define ROOTS_H

extern int SolveQuadric _PROTO((double[3], double[2]));
extern int SolveCubic _PROTO((double[4], double[3]));
extern int SolveQuartic _PROTO((double[5], double[4]));

#endif /* ROOTS_H */
