/* translate.h,v 1.1.1.1 1995/02/27 07:38:27 explorer Exp */

/*
 * Copyright (C) 1989, 1991, Craig E. Kolb
 * All rights reserved.
 *
 * This software may be freely copied, modified, and redistributed
 * provided that this copyright notice is preserved on all copies.
 *
 * You may not distribute this software, in whole or in part, as part of
 * any commercial product without the express consent of the authors.
 * 
 * There is no warranty or other guarantee of fitness of this software
 * for any purpose.  It is provided solely "as is".
 *
 */

#ifndef TRANSLATE_H
#define TRANSLATE_H

#define TransTranslateCreate()		TransCreate((TransRef)TranslateCreate(), TranslateMethods())

#define TransTranslateSetX(t, v)		TransAssoc(t, &((Translate *)t->tr)->x, v)
#define TransTranslateSetY(t, v)		TransAssoc(t, &((Translate *)t->tr)->y, v)
#define TransTranslateSetZ(t, v)		TransAssoc(t, &((Translate *)t->tr)->z, v)

typedef Vector Translate;

extern Translate *TranslateCreate _PROTO((void));
extern TransMethods *TranslateMethods _PROTO((void));
extern void TranslatePropagate _PROTO((Translate *, RSMatrix *, RSMatrix *));
void TranslationMatrix _PROTO((Float, Float, Float, RSMatrix *));

#endif /* TRANSLATE_H */
