/* acconfig.h,v 1.1.1.1 1995/02/27 07:38:23 explorer Exp */
/* Leave the following blank line there!!  Autoheader needs it.  */

/* define to use the Utah Raster toolkit */
#undef URT

/* define if your lex defined yylineno */
#undef HAVE_YYLINENO

/* define if you need to define sys_errlist[] */
#undef NEED_SYS_ERRLIST

/* Leave that blank line there!!  Autoheader needs it.
   If you're adding to this file, keep in mind:
   The entries are in sort -df order: alphabetical, case insensitive,
   ignoring punctuation (such as underscores).  */
