/* fbm.h,v 1.1.1.1 1995/02/27 07:38:41 explorer Exp */

/*
 * Copyright (C) 1989, 1991, Craig E. Kolb
 * All rights reserved.
 *
 * This software may be freely copied, modified, and redistributed
 * provided that this copyright notice is preserved on all copies.
 *
 * You may not distribute this software, in whole or in part, as part of
 * any commercial product without the express consent of the authors.
 *
 * There is no warranty or other guarantee of fitness of this software
 * for any purpose.  It is provided solely "as is".
 *
 */

#ifndef FBM_H
#define FBM_H

#define TextFBmCreate(o,s,h,l,n,t,m) TextCreate( \
			(TextRef)FBmCreate(o,s,h,l,n,t,m), FBmApply)
typedef struct {
  Float	beta;
  Float	omega;
  Float	lambda;
  Float	scale;
  Float	offset;
  Float	thresh;
  int	octaves;
  Color	*colormap;
} FBm;

extern FBm *FBmCreate _PROTO((Float, Float, Float, Float, int, Float, char *));
extern void FBmApply _PROTO((FBm *, Geom *, Ray *, Vector *, Vector *,
			     Vector *, Surface *));

#endif /* FBM_H */
